function [k,l,Nx, Ny, Qx, Qy, DnZ, c0, A] = hc_search( A, r, b )

% 2/20/2013
% Open-source implementation of the CoClusLSH algorithm for co-clustering a binary adjacency matrix.
% Please use for academic purposes and let the authors know if you have any questions.
% Authors: Tiantian Gao, Leman Akoglu (tiagao,leman@cs.stonybrook.edu)


% To run:

%Input: 

% Use [k,l,Nx, Ny, Qx, Qy, DnZ, c0, A] = hc_search( A, r, b )
% A: nxm binary adjacency matrix
% r,b: LSH parameters (we recommend r=20, b=5 for practitioners)

%Output:

%k: number of row clusters
%l: number of column clusters
%Nx: kx1 array of row-cluster sizes
%Ny: lx1 array of column-cluster sizes
%Qx: nx1 array of cluster labels for rows
%Qy: mx1 array of cluster labels for columns
% DnZ: kxl matrix of number of non-zeros in each block 
% c0: Final encoding cost
% A: input adjacency matrix with shuffled rows and columns

% The code generates an output visualization of the input matrix with the found clustering.


xperm = randperm(size(A,1));
yperm = randperm(size(A,2));   
A = A(xperm, yperm);


c0  = init_cost(A);
fprintf('Starting cost is %f bits\n', c0);
% pause


[k,l,Nx, Ny, Qx, Qy, DnZ] = init_perm( r, b, A );
[c0, trash] = cc_cost(k,l,Nx, Ny, DnZ);
fprintf('Cost after 1st iterations: %f\n', c0);
% size(DnZ)
%pause

  spyy=0;
% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% [Px, Py] = cc_perm (Qx, Qy);
% figure;
% if(~spyy)
% plot_binary_duo(A(Px,Py), 1, 1); 
% else
% spy(A(Px,Py));
% end
% [nx ny] = size(A);
% plot_cluster_grid(nx, ny, Nx, Ny);
% %pause
% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

cost_threshold = 10^-6;  % if cost change less than this, we stop
no_improvements = 0;
iter = 1;

r1=r; b1=b; C = 1.2;

while(1)

    [Ny1, Qy1,cluster_ids] = cos_col_hash2( round(r1/(C^(iter-1))), round(b1*(C^(iter-1))), k,l,Nx,Ny, Qy, DnZ );
    [l1, Ny1, DnZ1] = compute_Dnz2(Ny1,DnZ,cluster_ids);
    [c1, trash] = cc_cost(k,l1,Nx, Ny1, DnZ1);
    %pause

    if c0 - c1 < cost_threshold
        no_improvements = no_improvements + 1;
    else
        no_improvements = 0;
        c0 = c1;
        l=l1;
        Ny=Ny1;
        DnZ = DnZ1;
        Qy = Qy1;
        
        fprintf('Cost after col-cluster merge: %f\n', c0);
       
%         %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%         [Px, Py] = cc_perm (Qx, Qy);
%         figure;
%         if(~spyy)
%         plot_binary_duo(A(Px,Py), 1, 1); 
%         else
%         spy(A(Px,Py));
%         end
%         [nx ny] = size(A);
%         plot_cluster_grid(nx, ny, Nx, Ny);
%         %pause
%         %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    end

    [Nx1, Qx1, cluster_ids] = cos_col_hash2( round(r1/(C^(iter-1))), round(b1*(C^(iter-1))), l,k,Ny,Nx, Qx, DnZ' );
    [k1, Nx1, DnZp] = compute_Dnz2(Nx1,DnZ',cluster_ids);
    DnZ1 = DnZp';
    [c1, trash] = cc_cost(k1,l,Nx1, Ny, DnZ1);

    if c0 - c1 < cost_threshold
        no_improvements = no_improvements + 1;
    else
        no_improvements = 0;
        c0 = c1;
        k=k1;
        Nx=Nx1;
        DnZ = DnZ1;
        Qx = Qx1;
        
        fprintf('Cost after row-cluster merge: %f\n', c0);
        
%         %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%         [Px, Py] = cc_perm (Qx, Qy);
%         figure;
%         if(~spyy)
%         plot_binary_duo(A(Px,Py), 1, 1); 
%         else
%         spy(A(Px,Py));
%         end
%         [nx ny] = size(A);
%         plot_cluster_grid(nx, ny, Nx, Ny);
%         %pause
%         %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    end

    if no_improvements >= 2, break, end

   iter=iter+1;
end


% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% [Px, Py] = cc_perm (Qx, Qy);
% figure;
% if(~spyy)
% plot_binary_duo(A(Px,Py), 1, 1); 
% else
% spy(A(Px,Py));
% end
% [nx ny] = size(A);
% plot_cluster_grid(nx, ny, Nx, Ny);
% %pause
% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
fprintf('Last: %f\n', c0);
% Nx
% Ny

%pause
no_improvements = 0;
while(1)
    
    [Ny1, Qy1,cluster_ids] = cos_col_hash2( 1, 100, k,l,Nx,Ny, Qy, DnZ );
    [l1, Ny1, DnZ1] = compute_Dnz2(Ny1,DnZ,cluster_ids);
    [c1, trash] = cc_cost(k,l1,Nx, Ny1, DnZ1);
    %pause

    if c0 - c1 < cost_threshold
        no_improvements = no_improvements + 1;
    else
        no_improvements = 0;
        c0 = c1;
        l=l1;
        Ny=Ny1;
        DnZ = DnZ1;
        Qy = Qy1;
        
        fprintf('Cost after col-cluster merge: %f\n', c0);
       
%         %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%         [Px, Py] = cc_perm (Qx, Qy);
%         figure;
%         if(~spyy)
%         plot_binary_duo(A(Px,Py), 1, 1); 
%         else
%         spy(A(Px,Py));
%         end
%         [nx ny] = size(A);
%         plot_cluster_grid(nx, ny, Nx, Ny);
%         %pause
%         %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    end

    [Nx1, Qx1, cluster_ids] = cos_col_hash2( 1, 100, l,k,Ny,Nx, Qx, DnZ' );
    [k1, Nx1, DnZp] = compute_Dnz2(Nx1,DnZ',cluster_ids);
    DnZ1 = DnZp';
    [c1, trash] = cc_cost(k1,l,Nx1, Ny, DnZ1);

    if c0 - c1 < cost_threshold
        no_improvements = no_improvements + 1;
    else
        no_improvements = 0;
        c0 = c1;
        k=k1;
        Nx=Nx1;
        DnZ = DnZ1;
        Qx = Qx1;
        
        fprintf('Cost after row-cluster merge: %f\n', c0);
        
%         %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%         [Px, Py] = cc_perm (Qx, Qy);
%         figure;
%         if(~spyy)
%         plot_binary_duo(A(Px,Py), 1, 1); 
%         else
%         spy(A(Px,Py));
%         end
%         [nx ny] = size(A);
%         plot_cluster_grid(nx, ny, Nx, Ny);
%         %pause
%         %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    end
if no_improvements >= 2, break, end


end

fprintf('Final cost %f \n', c0)
%  Nx
%  Ny

        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        [Px, Py] = cc_perm (Qx, Qy);
        figure;
        if(~spyy)
        plot_binary_duo(A(Px,Py), 1, 1); 
        else
        spy(A(Px,Py));
        end
        [nx ny] = size(A);
        plot_cluster_grid(nx, ny, Nx, Ny);
        
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


end

